#
# Copyright (C) 2010 Martin Owens
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"""
Package for supplying default values and useful code for genetic wallpaper generation.
"""
import sys
import os
import random

from datetime import date, timedelta
from xml.dom.minidom import parse

# Uper and low bounds on scale and translate transforms.
# A, B, C, D, E, F - ScaleX, 0, 0, ScaleY, TranformX, TranformY
MINIMUM = [ 0.2, None, None, 0.2, 100, 100 ]
MAXIMUM = [ 3, None, None, 3, 1580, 950 ]


# Genetic mutation percentage
MUTATION = 0.02 # 2% of (max - min)
MIGRATION = 0.2 # 20% of target

def Elements(nodeList):
    """Create itorator for elements only"""
    for node in nodeList:
        if node.nodeType == 1:
            yield node

class SvgFileNotFound(IOError):
    pass

class SvgParseError(IOError):
    pass

class GeneticWallpaper(object):
    """Reprisent a genetic wallpaper svg"""
    def __init__(self, filename, plots=None):
        self.filename = filename
        (self.plot, self.weight) = self.todays_plot(plots)
        self.shapes = None

        if not os.path.exists(filename):
            raise SvgFileNotFound("Can not find SVG file: %s" % filename)
        try:
            self.xmldom = parse(filename)
        except:
            raise SvgParseError("Not a good XML file: %s" % filename)

        for group in self.xmldom.getElementsByTagName("g"):
             if group.getAttribute("id") == 'gen':
                 self.shapes = Elements(group.childNodes)
                 break
        if not self.shapes:
            raise SvgParseError("No generative groups in svg: %s." % filename)


    def nudge(self):
        """Nudge all shapes and save."""
        self.nudge_all_shapes()
        self.save()


    def save(self, target=None):
        """Save the svg back directly"""
        if not target:
             target = self.filename
        fhs = open(target, "w")
        fhs.write(self.xmldom.toxml())
        fhs.close()


    def generate_plot(self):
        """Return the plot for the loaded svg file"""
        result = []
        for shape in self.shapes:
            name = shape.getAttribute('id')
            result.append([name] + self.get_transform_matrix(shape))
        return result


    def get_transform_matrix(self, shape):
        """Returns a transformation matrix, corrects for translates"""
        attribute = shape.getAttribute('transform')
        kind = attribute.split('(')[0]
        matrix = attribute.split('(')[-1].split(')')[0].split(",")
        if kind == 'translate':
            # Force translate into a matrix transform
            matrix = [ '1', '0', '0', '1', matrix[0], matrix[1] ]
        elif len(matrix) != 6:
            # Used to catch errors
            matrix = [ '1', '0', '0', '1', '0', '0' ]
        return matrix


    def nudge_all_shapes(self):
        """Go through all shapes and nudge them all"""
        for shape in self.shapes:
            self.nudge_shape(shape)


    def nudge_shape(self, shape):
        """Nudge a single shape via it's transform"""
        name = shape.getAttribute('id') or ''
        plot = self.plot.get(name, None)
        matrix = self.get_transform_matrix(shape)

        for x in range(0, len(matrix)):
            if plot:
                matrix[x] = self.migrate(matrix[x], float(plot[x]), self.weight)
            else:
                matrix[x] = self.mutate(matrix[x], MINIMUM[x], MAXIMUM[x])
        tr = ",".join(matrix)
        shape.setAttribute('transform', "matrix(%s)" % tr)


    def mutate(self, element, emin, emax):
        """Mutate makes the shapes move randomly"""
        print "Mutating..."
        if emin and emax:
            element = float(element)
            amount = (float(random.randint(int(emin*100),
                int(emax*100))) / 100) * MUTATION
            if random.choice(('+', '-')) == '+':
                if element + amount <= emax:
                    element += amount
            else:
                if element - amount >= emin:
                    element -= amount
        return str(element)


    def migrate(self, element, target, weight=MIGRATION):
        """Migrate makes the shapes move towards predefined points"""
        print "Migrating..."
        element = float(element)
        diff = target - element
        if weight > 0 and weight <= 1:
            element = element + (diff * weight)
        return str(element)


    def todays_plot(self, filename):
        """Downloads a configuration file for convergence"""
        if not (filename and os.path.exists(filename)):
            return ({}, None)
        fhs = open(filename, "r")
        for line in fhs.readlines():
            cus = line.strip().split('|')
            dut = cus[0].split(',')
            start = date(int(dut[0]), int(dut[1]), int(dut[2]))
            steps = int(dut[3])
            end   = start + timedelta(steps)
            if date.today() >= start and date.today() < end:
                step_day = date.today() - start
                print "Plotting Day %s (%s)" % (str(step_day.days), str(1.0 / (steps - step_day.days)))
                if step_day.days < steps:
                    weight = 1.0 / (steps - step_day.days)
                    return (self.decode_plots(cus[1:]), weight)
        fhs.close()
        return ({}, None)


    def decode_plots(self, plots):
        """Return a hash/dict of plots and the objects they apply to"""
        result = {}
        for plot in plots:
            items = plot.split(',')
            result[items[0]] = items[1:]
        return result

